-- champion: Which champion to perform the attacks with. Can be Champion instance or just a champion index (not ordinal).
-- slot: Which item slot to use for the attacks (must be mainhand, 1, or offhand, 2).
-- special: true if the item's secondaryAction should be used, otherwise primaryAction will be used.
-- protection: protection of the monster, default 0
-- evasion: evasion of the monster, default 0. Set to "false" to make attacks always hit (damage against sleeping or frozen monster).
-- backstab: true to use backstabs (monster will face away from the party). Default false.
-- iterations: How many attacks to perform. More iterations = more accurate results, but will take longer. Default 1000.
-- summarize: prints result summary and creates in-game note. Default false.
--
-- Note: champion luck is not reset at the start of testing.
--
-- Returns: avg DPS, avg dam per attack, min dam, max dam, hit rate.
--
-- WARNING: If the slot is empty (unarmed attack), this will play the default swipe sound a number of times
-- equal to the number of iterations, so mute your audio if you do that.
-- Similarly, if you have a weapon with special on-hit effects like Dismantler you'll get lots of noise/light/particle
-- spam.
--
-- TODO: Finish implementing firearm,rangedattack,throwattack support
function testDamage(champion,slot,special,protection,evasion,backstab,summarize,iterations)
	if type(champion) == "number" then champion = party.party:getChampion(champion) end
	local item = champion:getItem(slot)
	local dx,dy
	local df = backstab and 2 or 0
	local attack
	iterations = iterations or 1000
	protection = protection or 0
	evasion = evasion == nil and 0 or evasion
	local startfood = champion:getFood()
	if item then
		attack = _getAttack(item,special)
		if attack then
			local cls = attack:getClass()
			if cls == "MeleeAttackComponent" or cls == "FirearmAttackComponent" then
				dx,dy = getForward(party.facing)
				if party.map:checkObstacle(party,party.facing) then
					print("testDamage needs an empty tile in front of the party")
					return
				end
			elseif cls == "RangedAttackComponent" or cls == "ThrowAttackComponent" then
				dx = 0
				dy = 0
			else
				print("unsupported attack class "..cls..", aborting")
				return
			end
		else
			print("item has no action, aborting")
			return
		end
	elseif special then
		print("requested special attack but champion is unarmed, aborting")
		return
	end
	local sound = attack and attack:getAttackSound()
	if attack then
		attack:setAttackSound("min_silence")
		if attack:getRepeatCount() and attack:getRepeatCount() ~= 1 then
			print("repeating attacks not yet implemented, aborting")
			return
		end
	end

	local mon = spawn("combat_test_monster",party.level,party.x+dx,party.y+dy,(party.facing+df)%4,party.elevation)
	mon.monster:setProtection(protection)
	if evasion == false then
		mon.monster:setCondition("frozen")
	else
		mon.monster:setEvasion(evasion)
	end
	_resetTestDamage()
	for i=1,iterations do
		champion:setFood(startfood)
		champion:attack(slot,special)
	end
	champion:setFood(startfood)
	if attack then attack:setAttackSound(sound) end
	local cooldown = _getTrueCooldown(champion,attack,special)
	local dpa = totalDam/iterations
	local dps = dpa/cooldown
	local hitRate = hits/iterations
	if summarize then
		local summary = string.format([[
Champion: %s   Weapon: %s%s
Monster protection: %d   Monster evasion: %s   Hit rate: %.2f%%
Damage per second: %.2f   Damage per attack: %.2f
Minimum damage: %d   Maximum damage: %d]],
champion:getName(),item and item:getUiName() or "Fists",special and " (Special attack)" or "",
protection,evasion == false and "Frozen" or tostring(evasion),hitRate*100,
dps,dpa,
lowestDam,highestDam)
		spawn("note").scrollitem:setScrollText(summary)
		local printFriendlySummary = summary:split("\n")
		for _,line in ipairs(printFriendlySummary) do print(line) end
	end
	mon:destroy()
	return dps,dpa,lowestDam,highestDam,hitRate
end

function _hit(damage)
	hits = hits+1
	totalDam = totalDam+damage
	highestDam = math.max(damage,highestDam)
	lowestDam = math.min(damage,lowestDam)
end

function _resetTestDamage()
	hits = 0
	totalDam = 0
	highestDam = 0
	lowestDam = math.huge
end

function _getAttack(item,special)
	local rval = item and (special and item:getSecondaryAction() or item:getPrimaryAction())
	if rval then
		return item.go:getComponent(rval)
	end
end

-- champion: Which champion to use to get strength, Aggressive, etc.
-- item: Item to get item info (attack power etc.) from. Pass a number to use the item in an inventory slot; if there is no item in the slot, unarmed melee damage will be calculated
-- special: true if the item's secondaryAction should be used, otherwise primaryAction will be used
-- protection: protection of the monster, default 0 -- TODO: NOT YET IMPLEMENTED, damage returned is for attacks on protectionless monsters
-- evasion: evasion of the monster, default 0. --TODO: NOT YET IMPLEMENTED, damage returned is for attacks on a sleeping/frozen monster
-- backstab: whether to compute for a backstab or not (note, backstab bonus will only apply if champion can actually backstab with the weapon)
-- critChance: champion's critical chance for the weapon (sorry but the scripting interface cannot get this easily). Default 5.
-- critMult: champion's critical multiplier for the weapon (like above, scripting interface can't get it well). Default 2.5 if non-special, 1.5 if special.
--
-- Note on computing damage: every non-crit attack in Grimrock has a 7% chance of having its damage divided by 2. Test this if you don't believe me. It's real. WTF.
-- I think I might have it in the wrong order...there's rounding in multiple places. Error should be very small, except at like, maybe attack power 3.
function computeDamage(champion,slot,special,protection,evasion,backstab,summarize,critChance,critMult)
	if type(champion) == "number" then champion = party.party:getChampion(champion) end
	critChance = (critChance or 5)/100
	critMult = critMult or special and 1.5 or 2.5
	local item = champion:getItem(slot)
	local attack = _getAttack(item,special)
	local low,high = _getTrueDamageRange(champion,item,attack)

	local cooldown = _getTrueCooldown(champion,attack,special)
	local dpa = (low+high)/2 -- Average damage of attack before crap, crits, etc.
	local crapChance = 0.07*(1-critChance) -- chance of damage being halved
	local nonCrapChance = 1-crapChance-critChance -- chance of neither crit nor crap

	-- Account for crap chance and criticals
	local dpa = dpa*nonCrapChance+math.max(1,dpa/2)*crapChance+dpa*critMult*critChance

	if backstab and item and (champion:hasTrait("backstab") and item:hasTrait("dagger") or champion:hasTrait("assassin") and item:hasTrait("light_weapon")) then
		dpa=dpa*3
		low=low*3
		high=high*3
	end

	local dual = champion:isDualWielding() or attack and attack:getClass() == "ThrowAttackComponent" and champion:hasTrait("double_throw")

	local dps = dpa/cooldown*(dual and 2 or 1)
	local lowestDam = math.floor(math.max(low*(critChance >= 100 and critMult or critChance > 0 and math.min(critMult,0.5) or 0.5),1))
	local highestDam = math.floor(math.max(high*(critChance >= 100 and critMult or critChance > 0 and math.max(critMult,1) or 1),1))
	local weaponName = string.format("%s%s%s",item and item:getUiName() or "Fists",dual and " (Dual)" or "",special and " (Special attack)" or "")
	if summarize then
		local summary = string.format([[
Champion: %s   Weapon: %s
Monster protection: 0   Monster evasion: Frozen   Hit rate: 100%%
Damage per second: %.2f   Damage per attack: %.2f
Minimum damage: %d   Maximum damage: %d]],
--protection,evasion == false and "Frozen" or tostring(evasion),hitRate*100,
champion:getName(),weaponName,
dps,dpa,
lowestDam,highestDam)
		spawn("note").scrollitem:setScrollText(summary)
		local printFriendlySummary = summary:split("\n")
		for _,line in ipairs(printFriendlySummary) do print(line) end
	end
	return dps,dpa,lowestDam,highestDam,hitRate,weaponName
end

-- TODO: account for crit chance and mult, specials
function rankWeapons(champion,protection,evasion,backstab,critChance,critMult)
	if type(champion) == "number" then champion = party.party:getChampion(champion) end
	local origItemL = champion:getItem(1)
	local origItemR = champion:getItem(2)
	if origItemL then champion:removeItemFromSlot(1) end
	if origItemR then champion:removeItemFromSlot(2) end

	local rankedList = {}
	for weapon,_ in pairs(Config.diagData.weapons) do
		local wp = spawn(weapon).item
		local wpr
		champion:insertItem(1,wp)
		if wp:hasTrait("light_weapon") and not (wp:hasTrait("two_handed") and not champion:hasTrait("two_handed_mastery")) then
			wpr = spawn(weapon).item
			champion:insertItem(2,wpr)
		end
		local dps,dpa,min,max,hitRate,weaponName = computeDamage(champion,1,false,protection,evasion,backstab,false,critChance,critMult)
		table.insert(rankedList,{dps=dps,name=weaponName})
		champion:removeItemFromSlot(1)
		if wpr then champion:removeItemFromSlot(2) end
	end
	table.sort(rankedList,function(a,b) return a.dps < b.dps end)

	if origItemL then champion:insertItem(1,origItemL) end
	if origItemR then champion:insertItem(2,origItemR) end

	print("Weapons from least to most damage per second:")
	for _,tab in ipairs(rankedList) do
		print(string.format("%.2f\t%s",tab.dps,tab.name))
	end
	print("WARNING: This function does not yet account for protection, evasion, or crit chance/mult bonuses on weapons.")
end

function _applyProtection(damage, protection, pierce)
	protection = protection-pierce
	if protection <= 0 then
		return damage
	elseif protection >= 1000 then
		return 0
	else
		return math.max(1,damage-math.floor((math.random()+0.5)*protection))
	end
end

-- Get the damage range of an attack, 
function _getDamageRange(basePower,statBonus)
	local baseMin = math.floor(basePower/2)
	local baseMax = math.floor(basePower*1.5)
	local statMin = math.floor(statBonus/2)
	local statMax = statBonus

	return baseMin+statMin,baseMax+statMax
end

function _getTrueDamageRange(champion,item,attack)
	local power = attack and attack:getAttackPower()*_getSkillAttackPowerMult(champion,item,attack) or _getUnarmedBasePower(champion) -- unarmed attack
	local stat
	local statBonus = 0
	if attack then
		stat = attack:getBaseDamageStat()
	else
		stat = "strength"
	end
	if stat then -- another distorted distribution; +1 damage for every even point, +0.5 damage for every odd point
		statBonus = math.floor(champion:getCurrentStat(stat))-10
	end
	if champion:hasTrait("aggressive") then statBonus=statBonus+4 end
	if champion:hasCondition("starving") then power = power/2 end
	local min,max = _getDamageRange(power,statBonus)
	return min,max
end

function _getSkillAttackPowerMult(champion,item,attack)
	local rval = 1
	if item then
		-- Skills increase *base* damage by 20% per level, except firearms because firearms aren't
		-- allowed to be good.
		if attack:getClass() == "MeleeAttackComponent" then
			-- If a weapon is both light and heavy, light weapons skill takes precedence.
			if item:hasTrait("light_weapon") then
				rval=rval+champion:getSkillLevel("light_weapons")*0.2
			elseif item:hasTrait("heavy_weapon") then
				rval=rval+champion:getSkillLevel("heavy_weapons")*0.2
			end
		elseif attack:getClass() == "RangedAttackComponent" and item:hasTrait("missile_weapon") then
			rval=rval+champion:getSkillLevel("missile_weapons")*0.2
		elseif attack:getClass() == "ThrowAttackComponent" and item:hasTrait("throwing_weapon") then
			rval=rval+champion:getSkillLevel("throwing")*0.2
		end
		-- Dual wielding reduces *base* damage
		if champion:isDualWielding() then
			if champion:hasTrait("rogue_dual_wield") then
				rval=rval*0.75
			else
				rval=rval*0.6
			end
		end
	end
	return rval
end

function _getTrueCooldown(champion,attack,special)
	local cooldownMod = 1
	local cooldownTraits = {
		quick=0.9,
		uncanny_speed=0.9,
		thunderstruck=0.5,
	}
	for trait,val in pairs(cooldownTraits) do
		if champion:hasTrait(trait) then cooldownMod = cooldownMod*val end
	end
	cooldownMod=cooldownMod*100/champion:getCurrentStat("cooldown_rate")
	local cooldown = attack and attack:getCooldown() or _getUnarmedBaseCooldown(champion)
	cooldown = cooldown*cooldownMod
	if attack and special then
		cooldown = cooldown+0.5+attack:getBuildup()*(attack:getClass() == "MeleeAttackComponent" and champion:hasTrait("melee_specialist") and 0.5 or 1)
	end
	return cooldown
end

function _getUnarmedBasePower(champion)
	return champion:hasCondition("bear_form") and 15 or 1
end
function _getUnarmedBaseCooldown(champion)
	return champion:hasCondition("bear_form") and 2.7 or 2.4
end

function dungeonReport()
--[[
	-- In order to prevent spam, item action and monster action/brain components are
	-- not included in this list because they will only appear on items or monsters,
	-- and ItemComponent and MonsterComponent are already unsafe.
	local UNSAFE_MINIMAL_COMPONENTS = {
		BeaconFurnaceControllerComponent = true,
		BlastComponent = true,
	--	BlindedMonsterComponent = true,
		BossFightComponent = true,
	--	BrainComponent = true,
	--	BurningMonsterComponent = true,
		ButtonComponent = true,
		CameraComponent = true,
		CameraShakeComponent = true,
	--	CastSpellComponent = true,
		ChestComponent = true,
		ClickableComponent = true,
		CloudSpellComponent = true,
	--	ContainerItemComponent = true,
		ControllerComponent = true, -- does not really have state information,
					-- but a controller on a minimalSaveState object
					-- is almost certainly a mistake
		CounterComponent = true,
	--	CrabBrainComponent = true,
	--	CraftPotionComponent = true,
		CrowControllerComponent = true,
	--	CrowernAttackComponent = true,
		CrystalComponent = true,
	--	CrystalShardItemComponent = true,
	--	DiggingToolComponent = true,
	-- DoorComponent is NOT included although it has state information, because it
	-- is often used for wall gratings, forest ruins walls, etc. that will never
	-- actually be opened.
		DynamicObstacleComponent = true, -- like ControllerComponent I am not
						-- sure this technically has meaningful
						-- state information, but is only useful
						-- for objects that require full saving
		EarthquakeComponent = true,
	--	EntangledMonsterComponent = true,
	--	EquipmentItemComponent = true,
	--	EyctopusBrainComponent = true,
	--	FireElementalBrainComponent = true,
	--	FirearmAttackComponent = true,	
		FloorTriggerComponent = true,
		ForceFieldComponent = true,
	--	FrozenMonsterComponent = true,
	--	GoromorgBrainComponent = true,
	--	GoromorgShieldComponent = true,
		GravityComponent = true,
		HealthComponent = true,
	--	HerderBigBrainComponent = true,
	--	HerderSmallBrainComponent = true,
	--	IceGuardianBrainComponent = true,
	--	IceLizardBrainComponent = true,
		IceShardsComponent = true,
	--	ItemActionComponent = true,
		ItemComponent = true,
		LeverComponent = true,
	--	LindwormBrainComponent = true,
	--	LindwormChargeComponent = true,
	--	LindwormFlyComponent = true,
		LockComponent = true,
	--	MagmaGolemBrainComponent = true,
		MapMarkerComponent = true,
	--	MeleeAttackComponent = true,
	--	MeleeBrainComponent = true,
	--	MimicCameraAnimationComponent = true,
	--	MonsterActionComponent = true,
	--	MonsterAttackComponent = true,
	--	MonsterChangeAltitudeComponent = true,
	--	MonsterChargeComponent = true,
		MonsterComponent = true,
	--	MonsterDropItemComponent = true,
	--	MonsterGroupComponent = true,
	--	MonsterJumpComponent = true,
	--	MonsterKnockbackComponent = true,
	--	MonsterLightCullerComponent = true,
	--	MonsterMoveAttackComponent = true,
	--	MonsterMoveComponent = true,
	--	MonsterOperateDeviceComponent = true,
	--	MonsterPickUpItemComponent = true,
	--	MonsterStealWeaponComponent = true,
	--	MonsterTurnComponent = true,
	--	MonsterWarpComponent = true,
	--	MosquitoSwarmBrainComponent = true,
	--	OgreBrainComponent = true,
		PartyComponent = true,
	-- For the same reason as DoorComponent, we don't include PitComponent.
		PoisonCloudAttackComponent = true,
	--	PoisonedMonsterComponent = true,
		PortalComponent = true, -- connectors at the very least
		ProjectileComponent = true,
		ProjectileImpactComponent = true,
		PullChainComponent = true,
		PushableBlockComponent = true,
		PushableBlockFloorComponent = true,
	--	RangedAttackComponent = true,
	--	RangedBrainComponent = true,
	--	RatlingBossBrainComponent = true,
	--	ReloadFirearmComponent = true,
	--	RopeToolComponent = true,
	--	RunePanelComponent = true,
		ScriptComponent = true,
		ScriptControllerComponent = true,
	--	ScrollItemComponent = true,
		SecretComponent = true,
	--	SkeletonArcherBrainComponent = true,
	--	SkeletonCommanderBrainComponent = true,
	--	SleepingMonsterComponent = true,
	--	SlimeBrainComponent = true,
		SmallFishControllerComponent = true,
		SocketComponent = true,
		SoundComponent = true,
		SpawnerComponent = true,
	--	SpellScrollItemComponent = true,
	--	StairsComponent = true, -- custom targets; ExitComponent doesn't support them so it's fine I think
		StatisticsComponent = true,
		StonePhilosopherControllerComponent = true,
	--	StunnedMonsterComponent = true,
		SurfaceComponent = true,
	--	SwarmBrainComponent = true,
		TeleporterComponent = true,
	--	TentacleBrainComponent = true,
	--	TentacleHideComponent = true,
		ThornWallComponent = true,
	--	ThrowAttackComponent = true,
		TileDamagerComponent = true,
		TimerComponent = true,
		TinyCritterControllerComponent = true,
	--	ToadBrainComponent = true,
		TorchHolderControllerComponent = true,
	--	TorchItemComponent = true,
	--	TricksterBrainComponent = true,
	--	TurtleBrainComponent = true,
	--	TwigrootBrainComponent = true,
	--	UggardianBrainComponent = true,
	--	UggardianFlamesComponent = true,
	--	UsableItemComponent = true,
	--	ViperRootBrainComponent = true,
		WallTextComponent = true,
		WallTriggerComponent = true,
	--	WargBrainComponent = true,
	--	WizardBrainComponent = true,
	--	XeloroidBrainComponent = true,
	--	ZarchtonBrainComponent = true,
	}]]

	-- Count objects and components.
	---------------------------------
	local minimalSaveTotal = 0
	local fullSaveTotal = 0
	local fullComponentTotal = 0
	local minimalComponentTotal = 0

	local fullObjectCounts = {}
	local minimalObjectCounts = {}

	local fullComponentCounts = {}
	local minimalComponentCounts = {}

	local objectNameIdLength = 0

	--local saveWarnings = {}

	local note = spawn("note")

	local function processObject(e)
		if not Config.diagData.definedObjects[e.name] then
			-- There is one object that the defineObject() shadow does not
			-- catch: starting_location. We don't care about it.
			return
		end
		local name = e.name
		objectNameIdLength = objectNameIdLength+e.id:len()+name:len()
		local def = Config.diagData.definedObjects[name]
		local minimal = def.minimalSaveState
		local loopGuard = {}
		while minimal == nil and def.baseObject and not loopGuard[def.name] do
			loopGuard[def.name] = true
			def = Config.diagData.definedObjects[def.baseObject]
			minimal = def.minimalSaveState
		end
		if minimal then
			minimalSaveTotal = minimalSaveTotal+1
			minimalObjectCounts[name] = minimalObjectCounts[name] and minimalObjectCounts[name]+1 or 1
		else
			fullSaveTotal = fullSaveTotal+1
			fullObjectCounts[name] = fullObjectCounts[name] and fullObjectCounts[name]+1 or 1
		end
		for _,c in e:componentIterator() do
			local cls = c:getClass()
			if minimal then
				minimalComponentTotal = minimalComponentTotal+1
				minimalComponentCounts[cls] = minimalComponentCounts[cls] and minimalComponentCounts[cls]+1 or 1
				--[[if UNSAFE_MINIMAL_COMPONENTS[cls] then
					if saveWarnings[name] then
						saveWarnings[name][cls] = true
					else
						saveWarnings[name] = {[cls] = true}
					end
				end]]
			else
				fullComponentTotal = fullComponentTotal+1
				fullComponentCounts[cls] = fullComponentCounts[cls] and fullComponentCounts[cls]+1 or 1
			end
			if cls == "ContainerItemComponent" then
				-- XXX: This will cause infinite recursion if a container contains itself, but the Grimrock engine
				-- already causes infinite recursion in that situation anyway.
				for slot=1,c:getCapacity() do
					local itm = c:getItem(slot)
					if itm then processObject(itm.go) end
				end
			elseif cls == "MonsterComponent" then
				-- I suppose theoretically you could make a monster contain itself too...
				-- We have to add a dummy item to the monster otherwise the next function
				-- from MonsterComponent:contents() crashes when called if the inventory
				-- is empty. We use the note. 
				c:addItem(note.item)
				for _,itm in c:contents() do
					if itm ~= note then processObject(itm.go) end
				end
				c:removeItem(note.item)
			end
		end
	end
	for l=1,Dungeon.getMaxLevels() do -- get all objects on the map and monster/container inventories...
		for e in Dungeon.getMap(l):allEntities() do
			processObject(e)
		end
	end

	for c=1,4 do -- ...plus all objects in champions' inventories.
		local champ = party.party:getChampion(c)
		for slot=1,ItemSlot.MaxSlots do
			local itm = champ:getItem(slot)
			if itm then processObject(itm.go) end
		end
	end

	-- Objects and components have been counted, now sort them.
	-- We used the object names and component classes as keys when counting,
	-- because if we used a straight array we would have to traverse it to
	-- find the object's entry, which is of course O(log(n)) compared to the
	-- O(1) of a table lookup. This does, however, mean that we must build
	-- a straight array now so we can sort it. Another option is to build
	-- the array while counting, but that isn't actually faster.
	-----------------------------------------------------------
	local cmpfunc = function(a,b) return a.c > b.c end

	local fullObjectsSorted = {}
	local minimalObjectsSorted = {}

	local fullComponentsSorted = {}
	local minimalComponentsSorted = {}

	for name,count in pairs(fullObjectCounts) do
		table.insert(fullObjectsSorted,{n=name,c=count})
	end
	for name,count in pairs(minimalObjectCounts) do
		table.insert(minimalObjectsSorted,{n=name,c=count})
	end
	for name,count in pairs(fullComponentCounts) do
		table.insert(fullComponentsSorted,{n=name,c=count})
	end
	for name,count in pairs(minimalComponentCounts) do
		table.insert(minimalComponentsSorted,{n=name,c=count})
	end

	table.sort(fullObjectsSorted,cmpfunc)
	table.sort(minimalObjectsSorted,cmpfunc)
	table.sort(fullComponentsSorted,cmpfunc)
	table.sort(minimalComponentsSorted,cmpfunc)

	local report = string.format([[
Total Objects: %d            Total Components: %d
Full Save State Objects: %d            Full Save State Components: %d
These objects and components will have their entire state saved.

Minimal Save State Objects: %d            Minimal Save State Components: %d
These objects will only have their name, id, level, x, y, facing, and elevation
saved. The components are not saved at all.

Total length of object names and ids: %d]],
fullSaveTotal+minimalSaveTotal,
fullComponentTotal+minimalComponentTotal,
fullSaveTotal,fullComponentTotal,
minimalSaveTotal,minimalComponentTotal,
objectNameIdLength)

	note.scrollitem:setScrollText(report..[[

For the full report including counts of each object and component, and warnings
about unsafe minimalSaveState usage, please check the game's terminal output.]])
	note.item:setUiName("Dungeon Report")
	if not getMouseItem() then setMouseItem(note.item) end

	print([[
================================================================================
Full save state objects:
================================================================================]])
	for _,t in ipairs(fullObjectsSorted) do
		print(t.c,t.n)
	end
	print([[
================================================================================
Minimal save state objects:
================================================================================]])
	for _,t in ipairs(minimalObjectsSorted) do
		print(t.c,t.n)
	end
	print([[
================================================================================
Full save state components:
================================================================================]])
	for _,t in ipairs(fullComponentsSorted) do
		print(t.c,t.n)
	end
	print([[
================================================================================
Minimal save state components:
================================================================================]])
	for _,t in ipairs(minimalComponentsSorted) do
		print(t.c,t.n)
	end
--[=[
	-- check if warnings are empty
	local warnings = false
	for _,i in pairs(saveWarnings) do
		warnings = true
		break
	end
	if warnings then
		print([[
================================================================================
WARNING: The following objects have components with extra state information, yet
you have defined them with minimalSaveState. Are you sure you know what you're
doing here?
================================================================================]])
		for name,t in pairs(saveWarnings) do
			local str = name..": "
			for component,_ in pairs(t) do
				str = str..component.." "
			end
			print(str)
		end
	end]=]
end
